package ru.dbotthepony.mc.otm.block.entity.cable

import net.minecraft.core.BlockPos
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.ForgeCapabilities
import ru.dbotthepony.mc.otm.SERVER_IS_LIVE
import ru.dbotthepony.mc.otm.block.CableBlock
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.math.BlockRotation
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.graph.GraphNode
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.util.Collections
import java.util.EnumMap

// after some thinking, team decided to settle with IC2's side (techreborn, gregtech, integrated dynamics*, pipez*, p2p tunnels, ...) of implementation,
// where cables have no residue capacitance, and never pull/push energy by themselves

// this allows simpler implementation and faster code, while also reducing possibility of duplication exploits

class EnergyCableBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryBlockEntity(MBlockEntities.ENERGY_CABLE, blockPos, blockState) {
	inner class CableSide(val side: RelativeSide) : IMatteryEnergyStorage {
		var isEnabled = true
			set(value) {
				field = value

				if (value) {
					node.graph.livelyNodes.add(node)
				}
			}

		init {
			check(side !in energySidesInternal)
			energySidesInternal[side] = this
			sides[side]!!.Cap(ForgeCapabilities.ENERGY, this)
		}

		val neighbour = sides[side]!!.trackEnergy()

		init {
			neighbour.addListener {
				if (isEnabled) {
					if (it.isPresent) {
						if (it.resolve().get() !is CableSide) {
							node.graph.livelyNodes.add(node)
						}

						ru.dbotthepony.mc.otm.onceServer {
							val newState = blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[blockRotation.side2Dir(side)]!!, true)

							if (newState !== blockState)
								level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
						}
					} else {
						ru.dbotthepony.mc.otm.onceServer {
							val newState = blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[blockRotation.side2Dir(side)]!!, false)

							if (newState !== blockState)
								level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
						}
					}
				}
			}
		}

		override fun extractEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
			return Decimal.ZERO
		}

		override fun receiveEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
			return node.graph.receiveEnergy(howMuch, simulate)
		}

		override var batteryLevel: Decimal
			get() = Decimal.ZERO
			set(value) {}

		override val maxBatteryLevel: Decimal get() = Decimal.ZERO
		override var energyFlow: FlowDirection = FlowDirection.BI_DIRECTIONAL
			private set

		override val canSetBatteryLevel: Boolean
			get() = false

		override val missingPower: Decimal
			get() = Decimal.POSITIVE_INFINITY
	}

	inner class Node : GraphNode<Node, EnergyCableGraph>(::EnergyCableGraph) {
		val sides get() = energySides

		override fun onNeighbour(link: Link) {
			if (link is DirectionLink) {
				val newState = blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[link.direction]!!, true)

				if (newState !== blockState && SERVER_IS_LIVE)
					level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
			}
		}

		override fun onUnNeighbour(link: Link) {
			if (link is DirectionLink) {
				val newState = blockState.setValue(CableBlock.MAPPING_CONNECTION_PROP[link.direction]!!, false)

				if (newState !== blockState && SERVER_IS_LIVE)
					level?.setBlock(blockPos, newState, Block.UPDATE_CLIENTS)
			}
		}
	}

	override val blockRotation: BlockRotation
		get() = BlockRotation.NORTH

	private val energySidesInternal = EnumMap<RelativeSide, CableSide>(RelativeSide::class.java)
	val energySides: Map<RelativeSide, CableSide> = Collections.unmodifiableMap(energySidesInternal)
	val node = Node()

	override fun setLevel(level: Level) {
		super.setLevel(level)
		node.discover(this, MatteryCapability.ENERGY_CABLE_NODE)
	}

	override fun setRemoved() {
		super.setRemoved()
		node.isValid = false
	}

	init {
		sides.keys.forEach { CableSide(it) }
		exposeGlobally(MatteryCapability.ENERGY_CABLE_NODE, node)
	}
}
